//Bare arch test created by Michal Pytlos, July 2015; header file

//This file is a part of Appendix C of: Pytlos, M. (2015), New tools for modelling soil-filled masonry arch bridges, 
//PhD thesis, Department of Civil and Structural Engineering, University of Sheffield, Sheffield.
//For the copyright details please search for the thesis on: http://etheses.whiterose.ac.uk/ 
//In no event will the author be held liable for any damages arising from the use of this computer program.

#ifndef MICHAL_ARCH_HEADER_H
#define MICHAL_ARCH_HEADER_H

//#define GUI

#include <iostream>
#include <sstream>
#include <fstream>
#include <math.h>       /* sin */
#include<ctime>
#ifdef _WIN32
	#include <windows.h>
#endif

#define myPI 3.14159265358979323846f
#define RADTODEG 57.295779513082320876f

#ifdef GUI
	#include <Testbed\Tests\michal_shape_definition.h>
#endif
#ifndef GUI
	#include "michal_shape_definition_no_gui.h"
	using namespace std;
#endif

//input
	//arch
float span = 20.0;//arch span
float spanToRise = 4;
float ringThickness = 1.2;
int noBricks = 3;
float myFriction = 0.6; //coefficient of friction for all bodies
float loadingPos = -0.25; //unit = arch span; relative to the center of the arch
float masonryUnitWeight = 20;// in kN/m2
	//other
float testGrav = -10.0;
float skinThickness = 0.0070; //aim for apporox 0.0070
float gravIncrTime = 10.0;//reach full gravity in this many seconds 
float maxLoadRate = 100;//in [kN/s]
float timeStepSize = 1.0/60.0;
int velocityIterations = 8;
int positionIterations = 3;	

//measure CPU time on Windows
double start_cpuWindows;
double stop_cpuWindows;
double cpuTimeWindows=0;

//bodies
const int noPBodies = 0;
const int noCBodies = 0;
const int noEBodies = 0;
int noQBodies;
int noTS = 0;// number of geometry entries to the outout files
const int maxNoBricks = 1000;
b2Body* myQuad[maxNoBricks+2]; //[0]=left abutment; [1]=right abutment; [2]=brick most to the left ... 
b2Body* myBrick[maxNoBricks];
//b2Body* myAbutment[2];//[0]=right abutment

//output
	std::ofstream outputStream[8];//[0]=testResults short file; [1]long file; [2]metaData; [3]=QGeometry; [4]=QGlobalX; [5]=QGlobalY; [6]=QRotation; [7]=identifiers;
	string outputDir1 ="../../../../output/bareArch/"; //windows only
	string outputDir2 ="../../../../output/bareArch/drawings/"; //windows only
	string fileName;
	string identifier;

void initalizeStreams(){//initialize streams for the outputs [drawing files only]
 
	string fileType[5] = {"metaData","QGeometry","QGlobalX","QGlobalY","QRotation"}; //this is asyName[] in retaining wall

	for (int i=0; i<5; ++i){//streams [2] to [6]
		fileName=outputDir2;
		fileName.append(identifier);
		fileName.append(fileType[i]);
		fileName.append(".txt");
		outputStream[i+2].open(fileName.c_str());
	}
}

void saveQGeom (){//save geometry of all quadrilaterals to QGeometry.txt;
	for (int i=0; i<noQBodies; ++i){
		b2Vec2 localCenterOfMassQ = myQuad[i]->GetLocalCenter();
		outputStream[3]<<x0Q[i]<<","<<x1Q[i]<<","<<x2Q[i]<<","<<x3Q[i]<<","<<localCenterOfMassQ.x<<","<<y0Q[i]<<","<<y1Q[i]<<","<<y2Q[i]<<","<<y3Q[i]<<","<<localCenterOfMassQ.y<<endl;
	}
}

void saveQSim(){//save simulation data (x,y,rotation) of all quadrilaterals to the relevant output files
	for (int i=0; i<noQBodies; ++i){	
		b2Vec2 bodyPosQ=myQuad[i]->GetPosition();
		float bodyRotQ=myQuad[i]->GetAngle(); //in radians
		outputStream[4]<<bodyPosQ.x<<",";
		outputStream[5]<<bodyPosQ.y<<",";
		outputStream[6]<<bodyRotQ<<",";
	}
	outputStream[4]<<endl;
	outputStream[5]<<endl;
	outputStream[6]<<endl;
}

void saveMeta(){//save meta data to metaData.txt
	outputStream[2]<<noQBodies<<","<<noPBodies<<","<<noCBodies<<","<<noEBodies<<","<<noTS;//write info to metaData file
}

//cast ray class
class myRayCast : public b2RayCastCallback 
{
public:

	myRayCast()
	{
		m_hit = false;
		myCount=0;
	}

	float32 ReportFixture(	b2Fixture* fixture, const b2Vec2& point,const b2Vec2& normal, float32 fraction){
		hitBody[myCount] = fixture->GetBody();
		m_fraction[myCount] = fraction;
		m_point[myCount] = point; //point on the body hit by the ray; in world coordinates
		m_hit = true;
		++myCount;
		
		return 1.0;
	}
	float m_fraction[maxNoP];
	b2Body* hitBody[maxNoP];
	int myCount;
	bool m_hit;
	b2Vec2 m_point[maxNoP];
};

//variables for contact listener
float myNormalImpulse[2];//[0]=left
float myTangentImpulse[2];
float myNormalForce[2];
float myTangentForce[2];
int numPoints; //number of collision contact points 
b2Vec2 myCollisionNormal;

//contact listener
class MyContactListener : public b2ContactListener
{
public:
 
	void PostSolve (b2Contact* contact,const b2ContactImpulse* impulse)
	{
		//identify colliding bodies
		const b2Body* bodyA = contact->GetFixtureA()->GetBody(); 
		const b2Body* bodyB = contact->GetFixtureB()->GetBody(); 

		//left abutment contact
		if (((bodyA == myQuad[2]) || (bodyB == myQuad[2])) && ((bodyA == myQuad[0]) || (bodyB == myQuad[0]))){	
			//check number of contact points
			numPoints = contact->GetManifold()->pointCount;

			if (numPoints == 2) {
				myNormalImpulse[0] = (impulse->normalImpulses[0] + impulse->normalImpulses[1]);
				myTangentImpulse[0] = (impulse->tangentImpulses[0] + impulse->tangentImpulses[1]);
			}

			//if there is only one contact point, ignore normalImpulses[1]
			if (numPoints == 1) {
				myNormalImpulse[0] = (impulse->normalImpulses[0]);
				myTangentImpulse[0] = (impulse->tangentImpulses[0]);
			}
		}

		//right abutment contact
		if (((bodyA == myQuad[noBricks-1+2]) || (bodyB == myQuad[noBricks-1+2])) && ((bodyA == myQuad[1]) || (bodyB == myQuad[1]))){	
			//check number of contact points
			numPoints = contact->GetManifold()->pointCount;

			if (numPoints == 2) {
				myNormalImpulse[1] = (impulse->normalImpulses[0] + impulse->normalImpulses[1]);
				myTangentImpulse[1] = (impulse->tangentImpulses[0] + impulse->tangentImpulses[1]);
			}

			//if there is only one contact point, ignore normalImpulses[1]
			if (numPoints == 1) {
				myNormalImpulse[1] = (impulse->normalImpulses[0]);
				myTangentImpulse[1] = (impulse->tangentImpulses[0]);
			}
		}
	}
};


b2Vec2 rotateVector (float angle, const b2Vec2& v) //angles in radians; rotate anticlockwise
{
	float s, c;
	s = sinf(angle);
	c = cosf(angle);

	//create rotation matrix	|r11 r12|
	//							|r21 r22|
		
	float r11, r12, r21, r22;
	r11 = c;
	r12 = -s;
	r21 = s;
	r22 = c;

	//multiply by column vector |r11 r12|	|x|
	//							|r21 r22|	|y|
		
	b2Vec2 v_g;
	v_g.x = r11*v.x + r12*v.y;
	v_g.y = r21*v.x + r22*v.y;

	return v_g;
};

#ifdef _WIN32
double cpuTimer() //Function to measure cpu time on Windows OS
{ 
	//structures to store times recorder by GetProcessTimes()
	FILETIME createTime; 
    FILETIME exitTime;
    FILETIME kernelTime;
    FILETIME userTime;

	GetProcessTimes (GetCurrentProcess( ), &createTime, &exitTime, &kernelTime, &userTime );//get timing	
	SYSTEMTIME userSystemTime;//structure to store user time
	FileTimeToSystemTime( &userTime, &userSystemTime );//covert times

	//return value in seconds
	return (double)userSystemTime.wHour * 3600.0 + (double)userSystemTime.wMinute * 60.0 + (double)userSystemTime.wSecond + (double)userSystemTime.wMilliseconds / 1000.0;
};
#endif
#endif

