//Biaxial compression test created by Michal Pytlos

//This file is a part of Appendix C of: Pytlos, M. (2015), New tools for modelling soil-filled masonry arch bridges, 
//PhD thesis, Department of Civil and Structural Engineering, University of Sheffield, Sheffield.
//For the copyright details please search for the thesis on: http://etheses.whiterose.ac.uk/ 
//In no event will the author be held liable for any damages arising from the use of this computer program.

#ifndef MICHAL_BIAXIAL_HEADER
#define MICHAL_BIAXIAL_HEADER

#include <sstream>
#include <fstream>
#include <math.h> /*tan; pow*/
#include <iostream>
#include <ctime> /* time_t, struct tm, time, localtime */

//#define GUI

#ifdef GUI
	#include <Testbed\Tests\michal_shape_definition.h>
#endif
#ifndef GUI
	#include "michal_shape_definition_no_gui.h"
	using namespace std;
#endif

#ifndef DEGTORAD /* this macro changes degrees to radians and vice-versa */
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

//Input variables and constants for michal_biaxial gui version
	//constants
	const int noQBodies = 1;//quadrilaterals
	const int noCBodies = 0;//circles
	const int noEBodies = 3;//edges
	const float capH = 0.2; //top cap height
	const float incrNo = 100.0;//number of pressure increments to reach the target confining pressure
	const float iniHtoW = 2.0; //initial sample height to width ratio; standard triaxial test: H=76mm W=38mm 
	const float noGravDecr = 1000.0;// number of gravity decrements to reach zero 
	const float noFriSteps = 100;// number of friction change steps
	float capConfVel = -0.001;

	//user inputs 
	float confPressure = 100; //in N/m; 
	float partIniFri = 1.0; //initial coefficient of friction used during deposition stage only
	float partFri = 0.3;
	float boundFri = 1.0;
	float partDensity = 2650;
	float capVel = -0.008; //velocity of the top cap
	float noPToWidth = 8;//number of particles to sample width ratio; this should be fixed for all tests
	unsigned int randSeed = 1; //don't use for now
	float depoGrav = -1.0; //gravity during sample deposition
	float HRHigh = 1.2; //height ratio high
	float strainLim = 0.1;// axial strain limit for shearing 
	float K_p = 0.005;//proportional gain
	float K_i = 0.0;//integral gain 
	float K_iFric = 0.0;//integral gain used during friction change
	float K_d = 0.0;//derivative gain

	//future user inputs
	const float particleRestitution = 0.0;
	const float circRL = 0.5; //circum radius low
	const float circRH = 0.5; //circum radius high
	const float irrF = 100.0; //irregularity factor

//variables for contact listener
	b2Body* myQuad[noQBodies];//myQuad[0] = top plate 
	b2Body* myParticle[maxNoP];
	b2Body* myEdge[noEBodies];
	int myCounter2, myCounter3;//used instead of 'i' to reduce CPU time spent on the loop. Counts particles in contact with the surface in a given time step 
	float particleImpulseB[maxNoP], particleImpulseT[maxNoP];
	int numPoints; //number of collision contact points 
	bool myReset;//bool to reset the contact

//variables accessible to functions
	std::ofstream outputStream[14];//[0]=testResults; [1]=metaData; [2]=QGeometry; [3]=QGlobalX; [4]=QGlobalY; [5]=QRotation; [6]=PGeometry; [7]=PGlobalX; [8]=PGlobalY; [9]=PRotation; [10]=PTotRot; [11]=EGeometry; [12]=identifiers;
#ifdef _WIN32
	string outputDir1 ="../../../../biaxial2/"; //windows
	string outputDir2 ="../../../../biaxial2/drawings/";
#endif
#ifdef __unix__
	string outputDir1 ="./MyResults/"; 
	string outputDir2 ="./MyResults/drawings/"; 
#endif
	string fileName;
	string identifier;
	float bodyTotRotP[maxNoP];//total rotation of a given particle
	int noTS = 0;// number of geometry entries to the outout files
	int noPBodies;

//contact listener
class MyContactListener : public b2ContactListener
{
public:
 
	void PostSolve (b2Contact* contact,const b2ContactImpulse* impulse)
	{
		//identify colliding bodies
		const b2Body* bodyA = contact->GetFixtureA()->GetBody(); 
		const b2Body* bodyB = contact->GetFixtureB()->GetBody(); 

		//get lateral force by adding impulses from all particles in contact with the dynamicWall
		for (int i=0; i<noPBodies; ++i) 
		{
			if (((bodyA == myParticle[i]) || (bodyB == myParticle[i])) && ((bodyA == myQuad[0]) || (bodyB == myQuad[0]))){	
				//check number of contact points
				numPoints = contact->GetManifold()->pointCount;

				if (numPoints == 2) {particleImpulseT[myCounter3] = (impulse->normalImpulses[0] + impulse->normalImpulses[1]);}

				//if there is only one contact point, ignore normalImpulses[1]
				if (numPoints == 1) {particleImpulseT[myCounter3] = (impulse->normalImpulses[0]);}

				++ myCounter3;
			}

			if (((bodyA == myParticle[i]) || (bodyB == myParticle[i])) && ((bodyA == myEdge[0]) || (bodyB == myEdge[0]))){	
				//check number of contact points
				numPoints = contact->GetManifold()->pointCount;

				if (numPoints == 2) {particleImpulseB[myCounter2] = (impulse->normalImpulses[0] + impulse->normalImpulses[1]);}

				//if there is only one contact point, ignore normalImpulses[1]
				if (numPoints == 1) {particleImpulseB[myCounter2] = (impulse->normalImpulses[0]);}

				++ myCounter2;
			}
		}

		if (myReset) {contact->ResetFriction();}//reset the contact for the new friction value to kick in
	}
};

//cast ray class
class myRayCast : public b2RayCastCallback 
{
public:

	myRayCast()
	{
		m_hit = false;
		myCount=0;
	}

	float32 ReportFixture(	b2Fixture* fixture, const b2Vec2& point,
		const b2Vec2& normal, float32 fraction)
	{
		hitBody[myCount] = fixture->GetBody();
		m_fraction[myCount] = fraction;
		m_point[myCount] = point; //point on the body hit by the ray; in world coordinates
		m_hit = true;
		++myCount;
		
		return 1.0;
	}
	float m_fraction[maxNoP];
	b2Body* hitBody[maxNoP];
	int myCount;
	bool m_hit;
	b2Vec2 m_point[maxNoP];
};

void createIdentifier(float param1, float param2, float param3, float param4, float param5, float param6, float param7, float param8, float param9, float param10, float param11, float param12){//create test identifier and append it to the identifiers.txt
		
	stringstream mySS;
	float identifierPart[12] = {param1, param2, param3, param4, param5, param6, param7, param8, param9, param10, param11, param12};

	for (int i=0; i<12; ++i){
		mySS<<identifierPart[i]; 
		identifier.append(mySS.str());
		identifier.append("_");
		mySS.clear();
		mySS.str(string());
	}

	identifier.append("biaxial_");

	//append identifier to the identifiers.txt
    fileName=outputDir2;
    fileName.append("Identifiers.txt");
    outputStream[13].open(fileName.c_str(), std::ios_base::app);
    outputStream[13]<<endl<<identifier;
    outputStream[13].close();
}

void initalizeStreams(){//initialize streams for the outputs
 
	string fileType[13] = {"results","metaData","QGeometry","QGlobalX","QGlobalY","QRotation","PGeometry","PGlobalX","PGlobalY","PRotation","PTotRot","EGeometry","isotropic"}; //this is asyName[] in retaining wall
	for (int i=0; i<13; ++i){
		if (i==0 || i ==12){fileName=outputDir1;}
		else {fileName=outputDir2;}
		fileName.append(identifier);
		if (i==12){//extra info in the iostropic file name 
			stringstream mySS;
			float identifierPart[3] = {K_p, K_i, K_d};
			for (int i=0; i<3; ++i){
				mySS<<identifierPart[i]; 
				fileName.append(mySS.str());
				fileName.append("_");
				mySS.clear();
				mySS.str(string());
			}
		}
		fileName.append(fileType[i]);
		fileName.append(".txt");
		outputStream[i].open(fileName.c_str());
	}
}

void saveQGeom (){//save geometry of all quadrilaterals to QGeometry.txt;
	for (int i=0; i<noQBodies; ++i){
		b2Vec2 localCenterOfMassQ = myQuad[i]->GetLocalCenter();
		outputStream[2]<<x0Q[i]<<","<<x1Q[i]<<","<<x2Q[i]<<","<<x3Q[i]<<","<<localCenterOfMassQ.x<<","<<y0Q[i]<<","<<y1Q[i]<<","<<y2Q[i]<<","<<y3Q[i]<<","<<localCenterOfMassQ.y<<endl;
	}
}

void saveEGeom(){//save geometry of all edges to EGeometry.txt;
	for (int i=0; i<noEBodies; ++i){
		outputStream[11]<<x0E[i]<<","<<x1E[i]<<","<<y0E[i]<<","<<y1E[i]<<endl;
	}
}

void savePGeom(){//save geometry of all dodecagon particles to PGeometry.txt;
	for (int i=0; i<noPBodies; ++i){
		b2Vec2 localCenterOfMassP = myParticle[i]->GetLocalCenter();
		outputStream[6]<<x0P[i]<<","<<x1P[i]<<","<<x2P[i]<<","<<x3P[i]<<","<<x4P[i]<<","<<x5P[i]<<","<<x6P[i]<<","<<x7P[i]<<","<<x8P[i]<<","<<x9P[i]<<","<<x10P[i]<<","<<x11P[i]<<","<<localCenterOfMassP.x<<","<<y0P[i]<<","<<y1P[i]<<","<<y2P[i]<<","<<y3P[i]<<","<<y4P[i]<<","<<y5P[i]<<","<<y6P[i]<<","<<y7P[i]<<","<<y8P[i]<<","<<y9P[i]<<","<<y10P[i]<<","<<y11P[i]<<","<<localCenterOfMassP.y<<endl;
	}
}

void saveQSim(){//save simulation data (x,y,rotation) of all quadrilaterals to the relevant output files
	for (int i=0; i<noQBodies; ++i){	
		b2Vec2 bodyPosQ=myQuad[i]->GetPosition();
		float bodyRotQ=myQuad[i]->GetAngle(); //in radians
		outputStream[3]<<bodyPosQ.x<<",";
		outputStream[4]<<bodyPosQ.y<<",";
		outputStream[5]<<bodyRotQ<<",";
	}
	outputStream[3]<<endl;
	outputStream[4]<<endl;
	outputStream[5]<<endl;
}

void savePSim(){
	for (int i=0; i<noPBodies; ++i){
		b2Vec2 bodyPosP = myParticle[i]->GetPosition();
		float bodyRotP = myParticle[i]->GetAngle();
		outputStream[7]<<bodyPosP.x<<",";
		outputStream[8]<<bodyPosP.y<<",";
		outputStream[9]<<bodyRotP<<",";
		outputStream[10]<<bodyTotRotP[i]<<",";
	}
	outputStream[7]<<endl;
	outputStream[8]<<endl;
	outputStream[9]<<endl;
	outputStream[10]<<endl;
}

void saveMeta(){//save meta data to metaData.txt
outputStream[1]<<noQBodies<<","<<noPBodies<<","<<noCBodies<<","<<noEBodies<<","<<noTS;//write info to metaData file
}

void destroyEdge(int ID){myEdge[ID]->GetWorld()->DestroyBody( myEdge[ID] );}//destroy this body safely

//variables for PIDcontrol()
const int maxNoPIDsteps = 500;
const int noAveSteps = 10; 
int PIDstep;
int AveStep;
float error;
float errorInteg;
float errorDeriv;
float errorAve;
float errorP[noAveSteps];
float errorD[noAveSteps];
float errorI[maxNoPIDsteps];

//This is not a textbook PID controller but it's roughly based on the same concept
void PIDcontrol (bool* success, float vConfForce, float timeStepSize, float aveAxialForce, int myTimeStep){
	++AveStep;
	++PIDstep;

	errorD[AveStep-1]=errorP[AveStep-1];//copy old value from errorP[]

	//proportional term
	error = (aveAxialForce-vConfForce)/vConfForce;//update normalized error
	errorP[AveStep-1]=error;//fill array
	float errSumP=0;
	for (int i=0; i<noAveSteps; ++i){errSumP+=errorP[i];}
	errorAve=errSumP/noAveSteps;	

	//derivative term 
	//slows the rate of change of the controller output
	float errSumD = 0;
	for (int i=0; i<noAveSteps; ++i){errSumD+=errorD[i];}
	errorDeriv=(errSumP-errSumD)/noAveSteps;
	
	//integral term
	//accelerates the movement of the process towards setpoint and eliminates the residual steady-state error
	errorI[PIDstep-1]=error;
	float errSumI = 0;
	for (int i=0; i<PIDstep; ++i){errSumI+=errorI[i];}//dynamic loop
	errorInteg = errSumI/PIDstep;

	//zero counters
	if (AveStep==noAveSteps){AveStep=0;}

	//modify capConfVel; depending on the amount of informtation available chose an appropriate method
	if (PIDstep>2*noAveSteps){capConfVel = K_p*errorAve+K_i*errorInteg*((float)PIDstep/(float)maxNoPIDsteps)+K_d*errorDeriv;}//(float)=cast into float
	else if (PIDstep>noAveSteps){capConfVel = K_p*errorAve+K_i*errorInteg*((float)PIDstep/(float)maxNoPIDsteps);}
	else {capConfVel = 0;}

	if (capConfVel>0.01){capConfVel=0.01;}//limit
	if (capConfVel<-0.01){capConfVel=-0.01;}
	myQuad[0]->SetLinearVelocity(b2Vec2(0, capConfVel));
	//0.01 = 1% slack on error; 1% slack on derivative 
	if ((PIDstep>noAveSteps*2 && (abs(errorAve)<0.01 && abs(errorDeriv)<0.01))|| PIDstep == maxNoPIDsteps){* success=true; PIDstep=0;}
	
	//labels for the output file
	if (myTimeStep==2){
			outputStream[12]<<"K_p,K_i,K_iFric,K_d"<<endl;
			outputStream[12]<<K_p<<","<<K_i<<","<<K_iFric<<","<<K_d<<endl<<endl;
			outputStream[12]<<"time,targetForce,measuredForce,errorAve,errorInteg,errorDeriv,capConfVel,PIDstep"<<endl;
	}
	outputStream[12]<<myTimeStep*timeStepSize<<","<<vConfForce<<","<<aveAxialForce<<","<<errorAve<<","<<errorInteg<<","<<errorDeriv<<","<<capConfVel<<","<<PIDstep<<endl;
}

#endif