//Particle shape definitions - created by Michal Pytlos June 2014 - no gui version

//This file is a part of Appendix C of: Pytlos, M. (2015), New tools for modelling soil-filled masonry arch bridges, 
//PhD thesis, Department of Civil and Structural Engineering, University of Sheffield, Sheffield.
//For the copyright details please search for the thesis on: http://etheses.whiterose.ac.uk/ 
//In no event will the author be held liable for any damages arising from the use of this computer program.

#include <Box2D/Box2D.h>

#ifndef DEGTORAD /* this macro changes degrees to radians and vice-versa */
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

#define	RAND_LIMIT	32767

/// Random floating point number in range [lo, hi]
inline float32 RandomFloat(float32 lo, float32 hi)
{
	float32 r = (float32)(std::rand() & (RAND_LIMIT));
	r /= RAND_LIMIT;
	r = (hi - lo) * r + lo;
	return r;
}

const int maxNoP = 20000;//maximum number of bodies of a given type
float x0P[maxNoP], x1P[maxNoP], x2P[maxNoP], x3P[maxNoP], x4P[maxNoP], x5P[maxNoP], x6P[maxNoP], x7P[maxNoP], x8P[maxNoP], x9P[maxNoP], x10P[maxNoP], x11P[maxNoP];
float y0P[maxNoP], y1P[maxNoP], y2P[maxNoP], y3P[maxNoP], y4P[maxNoP], y5P[maxNoP], y6P[maxNoP], y7P[maxNoP], y8P[maxNoP], y9P[maxNoP], y10P[maxNoP], y11P[maxNoP];
float x0Q[maxNoP], x1Q[maxNoP], x2Q[maxNoP], x3Q[maxNoP];
float y0Q[maxNoP], y1Q[maxNoP], y2Q[maxNoP], y3Q[maxNoP];
float x0E[maxNoP], x1E[maxNoP];
float y0E[maxNoP], y1E[maxNoP];

b2PolygonShape polyShape;
b2EdgeShape edgeShape;

//circumRadius Low [m] / circumRadius High [m] / HRHigh [1.0 = no elongation] /irrF [0-100]


void shapeDodecagon (float FcircRL, float FcircRH, float FirrF, float FHRHigh, int i)
{
	// dodecagon (12) shape definition
		// FcircRL = circum radius low
		// FcircRH = circum radius high
		// FirrF = irregularity factor [0-100]; 0 = regular shape
		// FHRHigh = height ratio high [1.0 = no elongation]
		// i = body id
	b2Vec2 vertices[12]; 

	float circR = RandomFloat(FcircRL, FcircRH); //calculate circum radius
	float HRLow = (FHRHigh-1)-FirrF/100*(FHRHigh-1)+1; //calculate low boundary for height ratio taking into accout irregularity factor
	float HR = RandomFloat (HRLow,FHRHigh); //calculate height ratio

		//1. set x0 and x6 to 0 
			x0P[i] = circR*sinf(0* DEGTORAD);
			x6P[i] = circR*sinf(-180* DEGTORAD);
			
		//2. set y3 and y9 to 0
			y3P[i] = circR*cosf(-90* DEGTORAD);
			y9P[i] = circR*cosf(-270* DEGTORAD);

		//3. set x3 and x9: distance between x3 and x8 must = 1.0 in order to keep specified height:width ratio; allowed x3: 1.0-1.25; allowed x9: 0.75-1.0
			x3P[i] = (1+FirrF/400)*circR*sinf(-90* DEGTORAD);
			x9P[i] = (1-FirrF/400)*circR*sinf(-270* DEGTORAD);

		//4. set y for all remaining vertices
			y0P[i] = HR*circR*cosf(0* DEGTORAD);
			y1P[i] = HR*circR*cosf(-30* DEGTORAD);
			y2P[i] = HR*circR*cosf(-60* DEGTORAD);
			y4P[i] = HR*circR*cosf(-120* DEGTORAD);
			y5P[i] = HR*circR*cosf(-150* DEGTORAD);
			y6P[i] = HR*circR*cosf(-180* DEGTORAD);
			y7P[i] = HR*circR*cosf(-210* DEGTORAD);
			y8P[i] = HR*circR*cosf(-240* DEGTORAD);
			y10P[i] = HR*circR*cosf(-300* DEGTORAD);
			y11P[i] = HR*circR*cosf(-330* DEGTORAD);

		//5. set x1 
			float x1_maxB = (x0P[i]-x3P[i])*(y1P[i]-y3P[i])/(y0P[i]-y3P[i])+x3P[i]; //max boundary to keep polygon convex
			float x1_minB = 0.7*x3P[i];//min boundary to keep polygon convex
			float x1_reg = circR*sinf(-30* DEGTORAD); //coordinate for a regular dodecagon
			float x1_max = (x1_maxB-x1_reg)*FirrF/100+x1_reg;//actual high bound taking into account irregularity factor
			float x1_min = (x1_minB-x1_reg)*FirrF/100+x1_reg;//actual low bound taking into account irregularity factor
			x1P[i] = RandomFloat(x1_min, x1_max);
			
		//6. set x2
			float x2_maxB = (x1P[i]-x3P[i])*(y2P[i]-y3P[i])/(y1P[i]-y3P[i])+x3P[i];//max boundary to keep polygon convex
			float y2_boundI = y1P[i]-(y0P[i]-y1P[i])*(x1P[i]-x3P[i])/(x0P[i]-x1P[i]);//calculate y of the intersection point for both min boundaries
			float x2_minB;
			if (y2_boundI >= y2P[i]) {x2_minB = 0.99*x3P[i];}//compare intersection point with y2 to decide which min bound to use
			if (y2_boundI < y2P[i]) {x2_minB = x1P[i]-(x0P[i]-x1P[i])*(y1P[i]-y2P[i])/(y0P[i]-y1P[i]);}
			float x2_reg = circR*sinf(-60* DEGTORAD);
			float x2_max = (x2_maxB-x2_reg)*FirrF/100+x2_reg;//actual high bound taking into account irregularity factor
			float x2_min = (x2_minB-x2_reg)*FirrF/100+x2_reg;//actual low bound taking into account irregularity factor
			x2P[i] = RandomFloat(x2_min, x2_max);
			
		//7. set x5
			float x5_maxB = (x6P[i]-x3P[i])*(y3P[i]-y5P[i])/(y3P[i]-y6P[i])+x3P[i]; 
			float x5_minB = 0.7*x3P[i];
			float x5_reg = circR*sinf(-150* DEGTORAD);
			float x5_max = (x5_maxB-x5_reg)*FirrF/100+x5_reg;
			float x5_min = (x5_minB-x5_reg)*FirrF/100+x5_reg;
			x5P[i] = RandomFloat(x5_min, x5_max);
			
		//8. set x4
			float x4_maxB = (x5P[i]-x3P[i])*(y3P[i]-y4P[i])/(y3P[i]-y5P[i])+x3P[i];
			float y4_boundI = y5P[i]+(y5P[i]-y6P[i])*(x5P[i]-x3P[i])/(x6P[i]-x5P[i]);
			float x4_minB;
			if (y4_boundI < y4P[i]) {x4_minB = 0.99*x3P[i];}
			if (y4_boundI >= y4P[i]) {x4_minB = x5P[i]-(x6P[i]-x5P[i])*(y4P[i]-y5P[i])/(y5P[i]-y6P[i]);}
			float x4_reg = circR*sinf(-120* DEGTORAD);
			float x4_max = (x4_maxB-x4_reg)*FirrF/100+x4_reg;
			float x4_min = (x4_minB-x4_reg)*FirrF/100+x4_reg;
			x4P[i] = RandomFloat(x4_min, x4_max);

		//9. set x7
			float x7_minB = (x9P[i]-x6P[i])*(y7P[i]-y6P[i])/(y9P[i]-y6P[i])+x6P[i]; 
			float x7_maxB = 0.7*x9P[i];
			float x7_reg = circR*sinf(-210* DEGTORAD);
			float x7_max = (x7_maxB-x7_reg)*FirrF/100+x7_reg;
			float x7_min = (x7_minB-x7_reg)*FirrF/100+x7_reg;
			x7P[i] = RandomFloat(x7_min, x7_max);

		//10. set x8
			float x8_minB = (x9P[i]-x7P[i])*(y8P[i]-y7P[i])/(y9P[i]-y7P[i])+x7P[i];
			float y8_boundI = y7P[i]+(y7P[i]-y6P[i])*(x9P[i]-x7P[i])/(x7P[i]-x6P[i]);
			float x8_maxB;
			if (y8_boundI < y8P[i]) {x8_maxB = 0.99*x9P[i];}
			if (y8_boundI >= y8P[i]) {x8_maxB = x6P[i]+(x7P[i]-x6P[i])*(y8P[i]-y6P[i])/(y7P[i]-y6P[i]);}
			float x8_reg = circR*sinf(-240* DEGTORAD);
			float x8_max = (x8_maxB-x8_reg)*FirrF/100+x8_reg;
			float x8_min = (x8_minB-x8_reg)*FirrF/100+x8_reg;
			x8P[i] = RandomFloat(x8_min, x8_max);
			
		//11. set x11
			float x11_minB = (x9P[i]-x0P[i])*(y0P[i]-y11P[i])/(y0P[i]-y9P[i])+x0P[i]; 
			float x11_maxB = 0.7*x9P[i];
			float x11_reg = circR*sinf(-330* DEGTORAD);
			float x11_max = (x11_maxB-x11_reg)*FirrF/100+x11_reg;
			float x11_min = (x11_minB-x11_reg)*FirrF/100+x11_reg;
			x11P[i] = RandomFloat(x11_min, x11_max);
			
		//12. set10
			float x10_minB = (x9P[i]-x11P[i])*(y11P[i]-y10P[i])/(y11P[i]-y9P[i])+x11P[i];
			float y10_boundI = y11P[i]-(y0P[i]-y11P[i])*(x9P[i]-x11P[i])/(x11P[i]-x0P[i]);
			float x10_maxB;
			if (y10_boundI >= y10P[i]) {x10_maxB = 0.99*x9P[i];}
			if (y10_boundI < y10P[i]) {x10_maxB = x0P[i]+(x11P[i]-x0P[i])*(y0P[i]-y10P[i])/(y0P[i]-y11P[i]);}
			float x10_reg = circR*sinf(-300* DEGTORAD);
			float x10_max = (x10_maxB-x10_reg)*FirrF/100+x10_reg;
			float x10_min = (x10_minB-x10_reg)*FirrF/100+x10_reg;
			x10P[i] = RandomFloat(x10_min, x10_max);

	vertices[0].Set(x0P[i], y0P[i]);
	vertices[1].Set(x1P[i], y1P[i]);
	vertices[2].Set(x2P[i], y2P[i]);
	vertices[3].Set(x3P[i], y3P[i]);
	vertices[4].Set(x4P[i], y4P[i]);
	vertices[5].Set(x5P[i], y5P[i]);
	vertices[6].Set(x6P[i], y6P[i]);
	vertices[7].Set(x7P[i], y7P[i]);
	vertices[8].Set(x8P[i], y8P[i]);
	vertices[9].Set(x9P[i], y9P[i]);
	vertices[10].Set(x10P[i], y10P[i]);
	vertices[11].Set(x11P[i], y11P[i]);
			
	polyShape.Set(vertices, 12);
}

void shapeROctagon (float FcircRL, float FcircRH, int i)
{
	// regular octagon (8) shape definition
	b2Vec2 vertices[8]; 

	float circR = RandomFloat(FcircRL, FcircRH); //calculate circum radius

	//x
		x0P[i] = circR*sinf(0* DEGTORAD);
		x1P[i] = circR*sinf(-45* DEGTORAD);
		x2P[i] = circR*sinf(-90* DEGTORAD);
		x3P[i] = circR*sinf(-135* DEGTORAD);
		x4P[i] = circR*sinf(-180* DEGTORAD);
		x5P[i] = circR*sinf(-225* DEGTORAD);
		x6P[i] = circR*sinf(-270* DEGTORAD);
		x7P[i] = circR*sinf(-315* DEGTORAD);

	//y 
		y0P[i] = circR*cosf(0* DEGTORAD);
		y1P[i] = circR*cosf(-45* DEGTORAD);
		y2P[i] = circR*cosf(-90* DEGTORAD);
		y3P[i] = circR*cosf(-135* DEGTORAD);
		y4P[i] = circR*cosf(-180* DEGTORAD);
		y5P[i] = circR*cosf(-225* DEGTORAD);
		y6P[i] = circR*cosf(-270* DEGTORAD);
		y7P[i] = circR*cosf(-315* DEGTORAD);

	vertices[0].Set(x0P[i], y0P[i]);
	vertices[1].Set(x1P[i], y1P[i]);
	vertices[2].Set(x2P[i], y2P[i]);
	vertices[3].Set(x3P[i], y3P[i]);
	vertices[4].Set(x4P[i], y4P[i]);
	vertices[5].Set(x5P[i], y5P[i]);
	vertices[6].Set(x6P[i], y6P[i]);
	vertices[7].Set(x7P[i], y7P[i]);
			
	polyShape.Set(vertices, 8);
}

void shapePentagon (float FcircRL, float FcircRH, int i)
{
	// regular pentagon (5) shape definition
	b2Vec2 vertices[5]; 

	float circR = RandomFloat(FcircRL, FcircRH); //calculate circum radius

	//set x
		x0P[i] = circR*sinf(0* DEGTORAD);
		x1P[i] = circR*sinf(-72* DEGTORAD);
		x2P[i] = circR*sinf(-144* DEGTORAD);
		x3P[i] = circR*sinf(-216* DEGTORAD);
		x4P[i] = circR*sinf(-288* DEGTORAD);

	//set y 
		y0P[i] = circR*cosf(0* DEGTORAD);
		y1P[i] = circR*cosf(-72* DEGTORAD);
		y2P[i] = circR*cosf(-144* DEGTORAD);
		y3P[i] = circR*cosf(-216* DEGTORAD);
		y4P[i] = circR*cosf(-288* DEGTORAD);

	vertices[0].Set(x0P[i], y0P[i]);
	vertices[1].Set(x1P[i], y1P[i]);
	vertices[2].Set(x2P[i], y2P[i]);
	vertices[3].Set(x3P[i], y3P[i]);
	vertices[4].Set(x4P[i], y4P[i]);
			
	polyShape.Set(vertices, 5);
}


void shapeOctagon (float vertX[], float vertY[], int bodyID)
{
	// normal octagon (8) shape definition
	b2Vec2 vertices[8]; 

	//x
		x0P[bodyID] = vertX[0];
		x1P[bodyID] = vertX[1];
		x2P[bodyID] = vertX[2];
		x3P[bodyID] = vertX[3];
		x4P[bodyID] = vertX[4];
		x5P[bodyID] = vertX[5];
		x6P[bodyID] = vertX[6];
		x7P[bodyID] = vertX[7];

	//y 
		y0P[bodyID] = vertY[0];
		y1P[bodyID] = vertY[1];
		y2P[bodyID] = vertY[2];
		y3P[bodyID] = vertY[3];
		y4P[bodyID] = vertY[4];
		y5P[bodyID] = vertY[5];
		y6P[bodyID] = vertY[6];
		y7P[bodyID] = vertY[7];

	vertices[0].Set(x0P[bodyID], y0P[bodyID]);
	vertices[1].Set(x1P[bodyID], y1P[bodyID]);
	vertices[2].Set(x2P[bodyID], y2P[bodyID]);
	vertices[3].Set(x3P[bodyID], y3P[bodyID]);
	vertices[4].Set(x4P[bodyID], y4P[bodyID]);
	vertices[5].Set(x5P[bodyID], y5P[bodyID]);
	vertices[6].Set(x6P[bodyID], y6P[bodyID]);
	vertices[7].Set(x7P[bodyID], y7P[bodyID]);
			
	polyShape.Set(vertices, 8);
}

void shapeQuad (float vertX[], float vertY[], int bodyID)
{
	b2Vec2 vertices[4]; 

	//x
		x0Q[bodyID] = vertX[0];
		x1Q[bodyID] = vertX[1];
		x2Q[bodyID] = vertX[2];
		x3Q[bodyID] = vertX[3];

	//y 
		y0Q[bodyID] = vertY[0];
		y1Q[bodyID] = vertY[1];
		y2Q[bodyID] = vertY[2];
		y3Q[bodyID] = vertY[3];

	vertices[0].Set(x0Q[bodyID], y0Q[bodyID]);
	vertices[1].Set(x1Q[bodyID], y1Q[bodyID]);
	vertices[2].Set(x2Q[bodyID], y2Q[bodyID]);
	vertices[3].Set(x3Q[bodyID], y3Q[bodyID]);
			
	polyShape.Set(vertices, 4);
}

void shapeEdge (float vertX[], float vertY[], int bodyID)
{
	//x
	x0E[bodyID] = vertX[0];
	x1E[bodyID] = vertX[1];
	//y
	y0E[bodyID] = vertY[0];
	y1E[bodyID] = vertY[1];

	edgeShape.Set(b2Vec2(x0E[bodyID], y0E[bodyID]), b2Vec2(x1E[bodyID], y1E[bodyID])); 
}